#ifdef DEBUG
	#include <iostream>
	using namespace std;
#endif

#include "FTPConnection.h"
#include "FTPDataConnection.h"

#include <qstringlist.h>
#include <qfile.h>
#include <qbuffer.h>
#include <qcstring.h>
#include <qstring.h>

struct GoodRet good_return_codes[] = 
{
	{"USER",	{230, 331, -1}	},
	{"PASS",	{230, 202, -1}	},
	{"CWD",		{250, -1}		},
	{"CDUP",	{250, -1}		},
	{"MKD",		{257, -1}		},
	{"RMD",		{250, -1}		},
	{"DELE",	{250, -1}		},
	{"TYPE",	{200, -1}		},
	{"QUIT",	{221, -1}		},
	{"PWD",		{257, -1}		},
	{"NOOP",	{200, -1}		},
	{"PASV",	{227, -1}		},
	{"RETR",	{125, 150, -1}	},
	{"eol", 	{-1}			}
};

FTPConnection::FTPConnection()
{
	socket = NULL;
	login_successful = false;
	busy = false;
	passive = false;
	delete_io = true;
	
	last_command = "";
	last_reply = "";
	last_reply_code = -1;

	QObject::connect(&data_connection, SIGNAL(transfer_complete()), this, SLOT(transfer_complete()));
}

FTPConnection::~FTPConnection()
{
	if(socket != NULL)
		disconnect();
}

bool FTPConnection::connect(const QString &host, Q_INT16 port, const QString &user, const QString &pass)
{
	if(connected())
	{
		disconnect();
	}
	login_successful = false;
	socket = new FTPRequestReplySocket();

	if(!socket->connect(host, port))
	{
		#ifdef DEBUG
			cout << "ERROR: Could not connect" << endl;
		#endif

		emit error("Could not connect");
		return false;
	}
	recv_message();
	
	if(last_reply_code != 220)
	{
		disconnect();
		return false;
	}
	
	if(!issue_command(tr("USER %1").arg(user)))
	{
		disconnect();
		return false;
	}

	if(last_reply_code == 331)
	{
		if(!issue_command(tr("PASS %1").arg(pass)))
		{
			disconnect();
			return false;
		}
	}

	login_successful = true;
	emit status_message("Connection established");
	
	#ifdef DEBUG
		cout << "STATUS: Connection established" << endl;
	#endif
		
	return true;
}

void FTPConnection::disconnect()
{
	login_successful = false;
	issue_command("QUIT");
	delete socket;
	socket = NULL;
}

void FTPConnection::transfer_complete()
{
	io_device->close();
	if(io_device != NULL && delete_io)
	{
		io_device->close();
		delete io_device;
		io_device = NULL;
	}
	recv_message();
	busy = false;
}

void FTPConnection::send_message(const QString &msg)
{
	#ifdef DEBUG
		cout << "SEND: " << msg << endl;;
	#endif
		
	last_command = msg;
	emit command_message(msg);
	socket->send_message(msg);
}

void FTPConnection::recv_message()
{
	last_reply = socket->recv_message();
	
	#ifdef DEBUG
		cout << "RECV: " << last_reply << endl;
	#endif
		
	emit server_message(last_reply);
	last_reply_code = last_reply.left(3).toInt();
}

bool FTPConnection::issue_command(const QString &command)
{
	if(socket == NULL)
		return false;

	cout << "sending message" << endl;
	send_message(command);
	cout << "message sent" << endl;
	recv_message();
	cout << "message received" << endl;

    if(!verify_return_code(command))
	{
    	#ifdef DEBUG
			cout << "ERROR: " << last_reply << endl;
		#endif
		
		error(last_reply);
		return false;
	}

	return true;
}

bool FTPConnection::verify_return_code(const QString &command)
{
	QString command_text;
	if(command.length() > 3)
	{
		command_text = command[3].isLetter() ? command.left(4) : command.left(3);
	}
	else
	{
		command_text = command.left(3);
	}

	for(int i = 0; good_return_codes[i].command != "eol"; ++i)
	{
		if(command_text == good_return_codes[i].command)
		{
			for(int j = 0; good_return_codes[i].good_codes[j] != -1; ++j)
			{
				if(last_reply_code == good_return_codes[i].good_codes[j])
					return true;
			}
		}
	}

	return false;
}

QString FTPConnection::get_path()
{
	QString dir = current_directory();
	if(dir[dir.length()-1] != '/')
		dir += '/';

	return dir;
}

bool FTPConnection::issue_path_command(const QString &com, const QString &path)
{
	if(path[0] == '/')
	{
		return issue_command(com + path);
	}
	else
	{
		QString dir = current_directory();
		if(dir[dir.length()-1] != '/')
			dir += '/';
		return issue_command(com + dir + path);
	}
}

void FTPConnection::change_directory(const QString &dir)
{
	QString com = "CWD ";
	issue_path_command(com, dir);
}

void FTPConnection::change_parent_directory()
{
	QString com = "CDUP";
	issue_command(com);
}

void FTPConnection::create_directory(const QString &dir)
{
	QString com = "MKD ";
	issue_path_command(com, dir);
}

void FTPConnection::delete_directory(const QString &dir)
{
	QString com = "RMD ";
	issue_path_command(com, dir);
}

QString FTPConnection::current_directory()
{
	QString com = "PWD";
	if(issue_command(com))
	{
		int first = last_reply.find('"');
		int last = last_reply.findRev('"');
		if(first == -1 || last == -1)
			return "";
		return last_reply.mid(first+1,last-(first+1));
	}

	return "";
}

void FTPConnection::delete_file(const QString &file)
{
	QString com = "DELE ";
	issue_path_command(com, file);
}

void FTPConnection::get_file(const QString &file)
{
	start_passive();
    QString path = get_path();
	send_message("RETR " + path + file);
	if(passive)
	{
		data_connection.establish_connection(data_address, data_port); 
        recv_message();
	}
	else
	{
		recv_message();
	}

cout << "Data connection established" << endl;
	if(!data_connection.is_open())
		return;
cout << "starting thread" << endl;

    QFile *f = new QFile(file);
	f->open(IO_WriteOnly);
	data_connection.set_iodevice(f);
	io_device = f;
	data_connection.start();
	busy = true;
}

void FTPConnection::put_file(const QString &file)
{

}

QStringList FTPConnection::dir_list(bool full /* =false */)
{
	set_type_ascii();
	start_passive();
		
		
	if(full)
		send_message("LIST");
	else
		send_message("NLST");

	if(passive)
	{
		data_connection.establish_connection(data_address, data_port); 
        recv_message();
	}
	else
	{
		recv_message();
	}
	
    QBuffer *buf = new QBuffer();
	delete_io = false;
	buf->open(IO_WriteOnly);
	data_connection.set_iodevice(buf);
	io_device = buf;
	data_connection.start();
	wait_for_transfer();
	
	buf->close();
	QString data(buf->buffer());
	delete buf;

	QStringList lst = QStringList::split('\n',data);
	return lst;
}

void FTPConnection::set_type_ascii()
{
	if(issue_command("TYPE A"))
	{
		data_connection.set_type(ASCII);
	}
}

void FTPConnection::set_type_binary()
{
	if(issue_command("TYPE I"))
	{
		data_connection.set_type(Image);
	}	
}

void FTPConnection::set_passive()
{
	passive = true;
}

void FTPConnection::start_passive()
{	
	issue_command("PASV");

	int left = last_reply.find('(');
	int right = last_reply.find(')');

	if(left == -1 || right == -1)
		return;

	QStringList bytes = QStringList::split(',',last_reply.mid(left+1,right-left-1));
	
#ifdef LOCAL
	data_address = "192.168.1.1";
#else
	data_address = bytes[0] + "." + bytes[1] + "." + bytes[2] + "." +bytes[3];
#endif
	data_port = (bytes[4].toInt()) * 256 + (bytes[5].toInt());
}

void FTPConnection::set_port(Q_INT16 port)
{
	QString addr = socket->address().toString();
	addr[addr.find('.')] = ',';
	addr[addr.find('.')] = ',';
	addr[addr.find('.')] = ',';
	addr += ',';

	QString s;
	addr += s.setNum(port / 256) + ',';
	addr += s.setNum(port % 256);
		
	issue_command("PORT " + addr);
	data_port = port;
	passive = false;
}

void FTPConnection::send_noop()
{
	issue_command("NOOP");
}
